#ifndef _SERVER_H
#define _SERVER_H

class MsgQ;

// player 0 calls this to initialize the server
// will create a socket on which the other players can connect
// to us
void InitServer();

// connect to the server
void ConnectServer(char *server);

void ServerGotData(MsgQ *q, int sock, int playerId);

#endif
