#ifndef _NETWORK_H
#define _NETWORK_H

class MsgQ;

enum { RestoreSaveFile, StartNewGame };

void HandleMessage(MsgQ *q, int recvSock, int info);
void HandleKey(int, int, int, char *);
void CivInit(int option, char *server);

#endif
