#include "misc.h"
#include "list.h"
#include <stdio.h>
#include <stdarg.h>

void AddMessage(List<charp> &l, char *format, ...)
{
  char str[256];
  va_list ap;
  va_start(ap, format);
  vsprintf(str, format, ap);
  va_end(ap);
  l.Insert(strdup(str));
}
