#ifndef _SAVE_FILE_H
#define _SAVE_FILE_H

#include "mytypes.h"

void CreateSaveFile();
void ReadSaveFile();

void WriteUChar(uchar);
void WriteUShort(ushort);
void WriteULong(ulong);
void WriteString(char *);
void WriteBytes(char *, int);

uchar ReadUChar();
ushort ReadUShort();
ulong ReadULong();
char *ReadString();
void ReadBytes(char *, int);

#endif
