#ifndef _MOVE_H
#define _MOVE_H

enum { PIECE_DIE, PIECE_MOVE, PIECE_CREATE, CITY_CREATE, CITY_DIE,
       CITY_SIZE, CITY_WORK_MAP, UNIT_UPDATE, DISCOVERY, WONDER,
       TERRAIN_CHANGE };

#include "mytypes.h"

struct PieceMove
{
  PieceMove(ulong Id, uchar Oper, ushort X, ushort Y, ushort Type = 0,
	    uchar Vet = 0, ulong City = 0)
    { id = Id; oper = Oper; x = X; y = Y; type = Type; city = City;
      veteran = Vet; }


  ulong id, city;
  ushort x, y, type;
  uchar oper, veteran;
  void NetworkOrder(); // converts to network ordering
  void HostOrder(); // converts to host ordering
};

#endif
