#ifndef _RIVERPICS_H
#define _RIVERPICS_H

char **WhichRiverBitmap(int numPoints, int a, int b, int c);

void DrawRiver(int x, int y, int riverPic);

char **DecodeRiverBitmap(int n);

int EncodeRiverBitmap(char **bitmap);

#endif
